package org.netkernel.ext.system.representation.impl;

import java.util.*;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.meta.IEndpointStateMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.util.MultiMap;
import org.netkernel.module.standard.endpoint.IStandardOverlay;
import org.netkernel.module.standard.endpoint.StandardTransportImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;

public class SpaceGraph
{
	private final Map<String,Edge[]> mEdgesOut = new HashMap<String, Edge[]>();
	private final Map<String,Edge[]> mEdgesIn = new HashMap<String, Edge[]>();
	private final String[] mRoots;
	private final String[] mFulcrums;
	
	public SpaceGraph(IRepDeployedModules aDeployedModules)
	{
		Object[] spaceHashes=aDeployedModules.getHierarchy().getValues("/modules/module//spaces/space/hash");
		
		//find all delegated logical endpoints
		int edgeCount=0;
		for (Object h: spaceHashes)
		{	String spaceHash=(String)h;
			IMetaRepresentation meta=aDeployedModules.metadataForHash(spaceHash);
			ISpaceMeta sm=(ISpaceMeta)meta;
			ISpaceElements elements=sm.getElements();
			ISpace lastSp=null;
			IEndpoint lastEp=null;
			Set<Edge> edges=new LinkedHashSet<Edge>();
			for (int i=0; i<elements.size(); i++)
			{	ISpace sp=elements.getDelegatedSpace(i);
				IEndpoint ep=elements.getPhysicalEndpoint(i);
				if (sp!=null)
				{	if (ep!=lastEp || sp!=lastSp)
					{	String epHash=aDeployedModules.hashForComponent(ep);
						String spHash=aDeployedModules.hashForComponent(sp);
						if (spHash!=null && epHash!=null)
						{	Edge edge=new Edge(epHash,spHash);
							edges.add(edge);
						}
					}
				}
				else
				{	//possible non-transparent overlay
					if (ep!=lastEp && ep instanceof IStandardOverlay && ep instanceof IEndpointStateMeta)
					{	Object dso=((IEndpointStateMeta)ep).getState().getValue(IStandardOverlay.STATE_DELEGATE_SPACE);
						if (dso instanceof ISpace)
						{	String epHash=aDeployedModules.hashForComponent(ep);
							String spHash=aDeployedModules.hashForComponent(dso);
							if (spHash!=null && epHash!=null)
							{	Edge edge=new Edge(epHash,spHash);
								edges.add(edge);
							}
						}
					}
				}
				lastSp=sp;
				lastEp=ep;
			}
			Edge[] edgeArray=new Edge[edges.size()];
			edgeArray=edges.toArray(edgeArray);
			edgeCount+=edgeArray.length;
			mEdgesOut.put(spaceHash, edgeArray);
		}
		
		//calculate references in now
		int multiExpectancy=edgeCount/mEdgesOut.size()*2;
		MultiMap refMap = new MultiMap(mEdgesOut.size(), multiExpectancy);
		for (Map.Entry<String, Edge[]> entry : mEdgesOut.entrySet())
		{	String spaceHash=entry.getKey();
			Edge[] edges=entry.getValue();
			for (Edge edge : edges)
			{	Edge refEdge=new Edge(edge.mPhysicalEndpointHash,spaceHash);
				refMap.put(edge.mSpaceHash, refEdge);
			}
		}
		//condense multimap
		for (Iterator i=refMap.keyIterator(); i.hasNext(); )
		{	String spaceHash=(String)i.next();
			List<Edge> edges=refMap.get(spaceHash);
			Edge[] edgeArray=new Edge[edges.size()];
			edgeArray=edges.toArray(edgeArray);
			mEdgesIn.put(spaceHash, edgeArray);
		}
		
		//find roots
		Set<String> roots=new HashSet<String>();
		for (Object h: spaceHashes)
		{	String spaceHash=(String)h;
			if (!mEdgesIn.containsKey(spaceHash))
			{	roots.add(spaceHash);
			}
		}
		mRoots=new String[roots.size()];
		roots.toArray(mRoots);
		
		//find fulcrums
		List<String> fulcrums=new ArrayList<String>();
		for (String rootHash:roots)
		{	ISpaceMeta meta=(ISpaceMeta)aDeployedModules.metadataForHash(rootHash);
			ISpaceElements elements=meta.getElements();
			IEndpoint lastEp=null;
			for (int i=0; i<elements.size(); i++)
			{	IEndpoint ep=elements.getPhysicalEndpoint(i);
				if (ep!=lastEp)
				{	lastEp=ep;
					String epHash=aDeployedModules.hashForComponent(ep);
					IMetaRepresentation epMeta=aDeployedModules.metadataForHash(epHash);
					Object isTransportObj=epMeta.getAdditionalFields().getValue(StandardTransportImpl.STATE_TRANSPORT_INDICATOR);
					boolean isTransport=isTransportObj instanceof Boolean && ((Boolean)isTransportObj).booleanValue();
					if (isTransport)
					{	fulcrums.add(rootHash);
						break;
					}
				}
			}
		}		
		mFulcrums=new String[fulcrums.size()];
		fulcrums.toArray(mFulcrums);
		

	}
	
	
	public Edge[] getEdgesOut(String aSpaceHash)
	{	return mEdgesOut.get(aSpaceHash);
	}
	
	public Edge[] getEdgesIn(String aSpaceHash)
	{	return mEdgesIn.get(aSpaceHash);
	}
	
	public static Set<String> getUniqueSpaceHashes(Edge[] aEdges)
	{	Set<String> result=new HashSet<String>();
		for (Edge edge: aEdges)
		{	result.add(edge.getSpaceHash());
		}
		return result;
	}
	
	public String[] getRoots()
	{	return mRoots;
	}
	
	public String[] getFulcrums()
	{	return mFulcrums;
	}
	
	public static class Edge
	{
		private final String mPhysicalEndpointHash;
		private final String mSpaceHash;
		
		public Edge(String aPhysicalEndpointHash, String aSpaceHash)
		{	mSpaceHash=aSpaceHash;
			mPhysicalEndpointHash=aPhysicalEndpointHash;
		}
		
		public boolean equals(Object aOther)
		{	if (aOther instanceof Edge)
			{	Edge other=(Edge)aOther;
				return other.mPhysicalEndpointHash.equals(this.mPhysicalEndpointHash) && other.mSpaceHash.equals(this.mSpaceHash);
			}
			else
			{	return false;
			}
		}
		
		public int hashCode()
		{	return this.mSpaceHash.hashCode();
		}
		
		
		public String getPhysicalEndpointHash()
		{	return mPhysicalEndpointHash;
		}
		
		public String getSpaceHash()
		{	return mSpaceHash;
		}
	}
}
